/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>          
#include <sys/stat.h> 
#include <cstring>              
#include <cerrno>
#include <fcntl.h>    
#include <sys/socket.h>
#include <string>

#include <src/investigation/donnees/serveur.h>

#include <src/investigation/donnees/liens.h>

#define TAILLE_BUFFER_LOCAL 1024
#define TAILLE_DONNEES_FLUX 4096 

using namespace Investigation::Donnees;

bool continue_server;

void sigterm(int)
{
	continue_server=false;
}

namespace Investigation
{
	namespace Donnees
	{
		namespace Outil
		{
			struct Memoire
			{
				Memoire()
				{
					res0 = 0;
					buffer_local=0;
				};

				~Memoire()
				{
					if(res0)
					{
						::freeaddrinfo(res0);
					}
					delete [] buffer_local;
				};
				void operator() (char *buffer)
				{
					delete buffer_local;
					buffer_local=buffer;
				};
				struct addrinfo *res0;
				char *buffer_local;
			};
		}
	}
}

bool Serveur::resolution_ip_port(const struct sockaddr* sock, const socklen_t longueur, std::string& ip, std::string& port)
{
        char host[TAILLE_DONNEES_FLUX+1];
        char serv[TAILLE_DONNEES_FLUX+1];
        bool resultat=::getnameinfo(sock,longueur,host,TAILLE_DONNEES_FLUX,serv,TAILLE_DONNEES_FLUX,NI_NUMERICHOST)==0;
        if(resultat)
        {
                ip = host;
                port = serv;
        }
        return resultat;
}

void Serveur::log_erreur(struct addrinfo *res, std::ostringstream& os, const std::string& fonction, int erreur)
{
        std::string ip;
        std::string port;
        os << std::endl << fonction << " ";
        if(Serveur::resolution_ip_port(res->ai_addr,res->ai_addrlen,ip,port))
        {
                os << "(" << ip << " : " << port << ")";
        }
        os << ": " << ::strerror(erreur);
}

	template<bool appelle_bind,bool appelle_listen, bool appelle_connect>
int Serveur::ouvrir(const int type, const int protocole, const std::string& ip, const std::string& port, const std::string& d)
{
	int fd;
	struct addrinfo hints, *res;
	int error;
	std::ostringstream details;
	details << Investigation::Global::Exception::details;
	Investigation::Donnees::Outil::Memoire memoire;

	::memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = type;
	hints.ai_protocol = 0;
	error = ::getaddrinfo(ip.c_str(), port.c_str(), &hints, &memoire.res0);
	if(error)
	{
		std::ostringstream details;
		details << Investigation::Global::Exception::details << std::endl << "getaddrinfo: " << ::strerror(errno);
		throw OuvertureFluxImpossible(d,details.str());
	}
	fd = -1;
	for (res = memoire.res0 ; res ; res = res->ai_next)
	{
		fd = ::socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (fd < 0)
		{
			Serveur::log_erreur(res,details,"socket",errno);
			continue;
		}

		if(appelle_bind)
		{
			if (::bind(fd, res->ai_addr, res->ai_addrlen) < 0)
			{
				Serveur::log_erreur(res,details,"bind",errno);
				::close(fd);
				fd = -1;
				continue;
			}
		}

		if(appelle_listen)
		{
			if(::listen(fd,100)<0)
			{
				Serveur::log_erreur(res,details,"listen",errno);
				::close(fd);
				fd = -1;
				continue;
			}
		}

		if(appelle_connect)
		{
			if (::connect(fd, res->ai_addr, res->ai_addrlen) < 0)
			{
				Serveur::log_erreur(res,details,"connect",errno);
				::close(fd);
				fd = -1;
				continue;
			}
		}
		break;
	}
	if (fd < 0)
	{
		throw OuvertureFluxImpossible(d,details.str());
	}
	return fd;
}


void Serveur::serveur(const bool tache_de_fond, const std::string& hotes, const std::string& port, const Investigation::Donnees::SiteSP& site)
{
	if(tache_de_fond)
	{
		pid_t pid=::fork();
		if(pid<0)
		{
			throw ImpossibleDePasserEnTacheDeFond();
		}
		if(pid!=0)
			return;
		::close(STDIN_FILENO);
		::close(STDOUT_FILENO);
	}
	::signal(SIGTERM,sigterm);
	Investigation::Donnees::Outil::Memoire memoire;
	int fdattente = Serveur::ouvrir<true,true,false>(SOCK_STREAM,0,hotes,port,"connector");
	Investigation::Donnees::Liens liens;
	continue_server=true;
	while(continue_server)
	{
		struct sockaddr client;
		socklen_t longueur = sizeof(client);
		int fd;
		errno=0;
		do
		{
			fd = ::accept(fdattente,&client,&longueur);
		}
		while(errno==EAGAIN);
		if (fd < 0)
		{
			if(errno!=EINTR)
			{
				std::ostringstream details;
				details << Investigation::Global::Exception::details << std::endl << "accept: " << ::strerror(errno); 
				throw AttenteFluxImpossible("connector",details.str());
			}
		}
		else
		{
			std::stringstream _buffer;
			int taille = 1;
			int flags = ::fcntl(fd, F_GETFL, 0);
			::fcntl(fd, F_SETFL, flags & (~O_NONBLOCK));
			char buffer_local[TAILLE_BUFFER_LOCAL];
			while((taille=::read(fd,buffer_local,sizeof(buffer_local)-1))>0)
			{
				if(std::find(buffer_local,buffer_local+taille,'\n')!=(buffer_local+taille))
					break;
				_buffer.write(buffer_local,taille);
			}
			_buffer.write(buffer_local,taille);
			std::string requete_http;
			std::getline(_buffer,requete_http);
			if(requete_http.substr(0,4)!="GET ")
			{
				std::ostringstream oss;
				oss << HttpBadRequest::defaut(site->_titre,"/style.css");
				if(::write(fd,oss.str().c_str(),oss.str().size())<0)
				{
				}
			}
			else
			{
				requete_http = requete_http.substr(4);
				requete_http = std::string(requete_http.cbegin(),std::find(requete_http.cbegin(),requete_http.cend(),' '));
				std::string element_http;
				std::pair<std::string,std::string> parametre_http;
				auto it_param = std::find(requete_http.cbegin(),requete_http.cend(),'?');
				element_http = std::string(requete_http.cbegin(),it_param);
				if(it_param!=requete_http.end())
				{
					++it_param;
					requete_http=std::string(it_param,requete_http.cend());
					auto it_valeur = std::find(requete_http.cbegin(),requete_http.cend(),'=');
					parametre_http.first=std::string(requete_http.cbegin(),it_valeur);
					if(it_valeur!=requete_http.cend())
					{
						++it_valeur;
						parametre_http.second=std::string(it_valeur,requete_http.cend());
					}
				}
				if((element_http=="/investigate.html") or (element_http=="/"))
				{
					if((not parametre_http.first.empty()) and (parametre_http.first!="q"))
					{
						std::ostringstream oss;
						oss << HttpBadRequest::defaut(site->_titre,"/style.css");
						if(::write(fd,oss.str().c_str(),oss.str().size())<0)
						{
						}
					}
					Requete requete = liens[parametre_http.second];
					if(not tache_de_fond)
					{
						std::cout << "Link [" << parametre_http.second << "] requested:" << requete << std::endl;
					}
					Investigation::Donnees::Html::Html reponse = (*site)((*site)(requete),"/investigate.html?q=",requete,liens);
					std::ostringstream oss;
					oss << HttpOK(reponse);
					if(::write(fd,oss.str().c_str(),oss.str().size())<0)
					{
					}
				}
				else if(element_http=="/style.css")
				{
					std::ostringstream oss;
					oss << HttpOK("text/css ; charset=UTF-8",site->_style);
					if(::write(fd,oss.str().c_str(),oss.str().size())<0)
					{
					}
				}	
				else
				{	
					std::ostringstream oss;
					oss << HttpNotFound::defaut(site->_titre,"/style.css");
					if(::write(fd,oss.str().c_str(),oss.str().size())<0)
					{
					}
				}
			}
			::close(fd);
		}
	}
	::close(fdattente);
}
